/*
    DoubleSimple: DoubleLong.m
    Copyright (C) 2003  Macaulay Institute

    This file is part of DoubleSimple, a series of classes for implementing
    various remedies to floating point arithmetic errors.

    DoubleSimple is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    DoubleSimple is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/
/*

Implementation for the DoubleLong class.

*/

#import "DoubleLong.h"

#import <stdlib.h>
#import <math.h>
#import <float.h>
#import <errno.h>
#import <string.h>
#import <stdlib.h>
#import <objc/objc-api.h>
#import <ctype.h>
#ifdef INTEL
#  import <CIieeefp.h>
static fp_pctl save_pctl;
#endif

#define NZ_ACC 40		// Number of zeros back from the 'e' in a
				// floating point number printed using %.*e
				// format required to be sure that the
				// floating point number is printed to
				// maximum accuracy

@implementation DoubleLong

+newUp {
  DoubleLong *obj = [super newUp];

#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_EXT);
#endif

  obj->ld = (long double)obj->d;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif

  return obj;
}

-copy {
  DoubleLong *obj = [super copy];

  [obj lcset: ld];

  return obj;
}

-set: (DoubleSimple *)value {
  [super set: value];
  if([value isKindOf: [DoubleLong class]]) {
    ld = [(DoubleLong *)value lget];
  }
  else {
#ifdef INTEL
    save_pctl = fpsetprecision(FP_PC_EXT);
#endif

    ld = (long double)d;

#ifdef INTEL
    fpsetprecision(save_pctl);
#endif
  }
  return self;
}

-cset: (double)value {
  [super cset: value];
#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_EXT);
#endif

  ld = (long double)d;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif
  return self;
}

-lcset: (long double)value {
  ld = value;
#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_DBL);
#endif

  d = (double)ld;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif
  return self;
}

-setStr: (char *)str {
  ld = [self lcheckConst: str];
#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_DBL);
#endif

  d = (double)ld;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif
  return self;
}

-(double)get {
  double dd;
#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_DBL);
#endif

  dd = (double)ld;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif

  return dd;
}

-(long double)lget {
  return ld;
}


-read: (FILE *)fp {
#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_EXT);
#endif

  fscanf(fp, "%Lf", &ld);

#ifdef INTEL
  fpsetprecision(FP_PC_DBL);
#endif

  d = (double)ld;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif
  return self;
}

-write: (FILE *)fp {
#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_EXT);
#endif

  fprintf(fp, "%Lg", ld);

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif
  return self;
}

-write: (FILE *)fp sf: (int)sigfig {
#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_EXT);
#endif

  fprintf(fp, "%.*Le", sigfig - 1, ld);

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif
  return self;
}

-writeAccurately: (FILE *)fp {
  char *buf, tbuf[3];
  int dp;
  int nz;

#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_EXT);
#endif

  dp = NZ_ACC;
  do {
    char *p, *q;
    int len = snprintf(tbuf, 2, "%.*Le", dp, ld);
				// Use snprintf and a dummy buffer to get the
				// size of buffer needed to print the floating
				// point number to the required number of
				// decimal places
    if(len < 0) {
      perror("Error in call to snprintf");
      abort();
    }

    buf = (char *)malloc(len + 1);
				// Allocate memory for the buffer
    if(buf == NULL) {
      perror("Memory allocation");
      abort();
    }

    sprintf(buf, "%.*Le", dp, ld);
				// Print the floating point number to the
				// buffer in the format X.YYYYYY0000eZZZ
    p = strchr(buf, (int)'e');
				// Find the 'e' in the buffer.
    if(p == NULL) {		// If we can't find the 'e' then this is some
				// funny number like NaN or Inf
      fprintf(fp, "%Le", ld);
      free(buf);

#ifdef INTEL
      fpsetprecision(save_pctl);
#endif

      return self;
    }
    q = p;			// Remember where it is...
      
    p--;			// Look back from the 'e' 
    nz = 0;			// and count the number
    while((*p) == '0') {	// of consecutive '0's
      nz++;
      p--;
    } 				// Right now, this should be the state of play:
				// X.YYYYYY0000eZZZ
				// ^buf   ^p   ^q
				// If the number is like X.0000000000eZZZ,
				// then p points here:    ^p.

    if(nz >= NZ_ACC) {		// We've got enough zeros before the 'e' to
				// ensure that maximum accuracy has been
				// achieved (according to the NZ_ACC parameter)
      if((*p) != '.') {
	p++;			// p didn't point to the decimal place, so
				// shift it right one place to point to the
				// first in a long line of zeros.
      }
      (*p) = '\0';		// Overwrite the decimal point or the first
				// in a long line of zeros with a string
				// terminator. Now we have (using # for the
				// terminator):
				// X.YYYYYY#000eZZZ# or X#0000000000eZZZ#
				// ^buf    ^p  ^q       ^buf        ^q
      fprintf(fp, "%s%s", buf, q);
				// So, printing buf then q gives:
				// X.YYYYYYeZZZ      or XeZZZ
    }
    else {
      dp += NZ_ACC;		// Increment the number of decimal places to
				// print by the required accuracy
    }

    free(buf);			// Free the buffer
  } while(nz < NZ_ACC);

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif
  return self;
}


-abs {
#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_EXT);
#endif

  if(ld < 0.0) ld *= -1.0L;

#ifdef INTEL
  fpsetprecision(FP_PC_DBL);
#endif

  d = (double)ld;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif
  return self;
}


-add: (DoubleSimple *)value {
#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_EXT);
#endif

  if([value isKindOf: [DoubleLong class]]) {
    ld += [(DoubleLong *)value lget];
  }
  else {
    ld += (long double)[value get];
  }

#ifdef INTEL
  fpsetprecision(FP_PC_DBL);
#endif

  d = (double)ld;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif
  return self;
}

-sub: (DoubleSimple *)value {
#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_EXT);
#endif

  if([value isKindOf: [DoubleLong class]]) {
    ld -= [(DoubleLong *)value lget];
  }
  else {
    ld -= (long double)[value get];
  }

#ifdef INTEL
  fpsetprecision(FP_PC_DBL);
#endif

  d = (double)ld;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif
  return self;
}

-mul: (DoubleSimple *)value {
#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_EXT);
#endif

  if([value isKindOf: [DoubleLong class]]) {
    ld *= [(DoubleLong *)value lget];
  }
  else {
    ld *= (long double)[value get];
  }

#ifdef INTEL
  fpsetprecision(FP_PC_DBL);
#endif

  d = (double)ld;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif
  return self;
}

-div: (DoubleSimple *)value {
#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_EXT);
#endif

  if([value isKindOf: [DoubleLong class]]) {
    ld /= [(DoubleLong *)value lget];
  }
  else {
    ld /= (long double)[value get];
  }

#ifdef INTEL
  fpsetprecision(FP_PC_DBL);
#endif

  d = (double)ld;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif
  return self;
}


-iadd: (int)value {
#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_EXT);
#endif

  ld += (long double)value;

#ifdef INTEL
  fpsetprecision(FP_PC_DBL);
#endif

  d = (double)ld;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif
  return self;
}

-isub: (int)value {
#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_EXT);
#endif

  ld -= (long double)value;

#ifdef INTEL
  fpsetprecision(FP_PC_DBL);
#endif

  d = (double)ld;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif
  return self;
}

-imul: (int)value {
#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_EXT);
#endif

  ld *= (long double)value;

#ifdef INTEL
  fpsetprecision(FP_PC_DBL);
#endif

  d = (double)ld;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif
  return self;
}

-idiv: (int)value {
#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_EXT);
#endif

  ld /= (long double)value;

#ifdef INTEL
  fpsetprecision(FP_PC_DBL);
#endif

  d = (double)ld;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif
  return self;
}


-cadd: (double)value {
#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_EXT);
#endif

  ld += (long double)value;

#ifdef INTEL
  fpsetprecision(FP_PC_DBL);
#endif

  d = (double)ld;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif
  return self;
}

-csub: (double)value {
#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_EXT);
#endif

  ld -= (long double)value;

#ifdef INTEL
  fpsetprecision(FP_PC_DBL);
#endif

  d = (double)ld;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif
  return self;
}

-cmul: (double)value {
#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_EXT);
#endif

  ld *= (long double)value;

#ifdef INTEL
  fpsetprecision(FP_PC_DBL);
#endif

  d = (double)ld;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif
  return self;
}

-cdiv: (double)value {
#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_EXT);
#endif

  ld /= (long double)value;

#ifdef INTEL
  fpsetprecision(FP_PC_DBL);
#endif

  d = (double)ld;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif
  return self;
}


-lcadd: (long double)value {
#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_EXT);
#endif

  ld += value;

#ifdef INTEL
  fpsetprecision(FP_PC_DBL);
#endif

  d = (double)ld;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif
  return self;
}

-lcsub: (long double)value {
#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_EXT);
#endif

  ld -= value;

#ifdef INTEL
  fpsetprecision(FP_PC_DBL);
#endif

  d = (double)ld;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif
  return self;
}

-lcmul: (long double)value {
#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_EXT);
#endif

  ld *= value;

#ifdef INTEL
  fpsetprecision(FP_PC_DBL);
#endif

  d = (double)ld;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif
  return self;
}

-lcdiv: (long double)value {
#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_EXT);
#endif

  ld /= value;

#ifdef INTEL
  fpsetprecision(FP_PC_DBL);
#endif

  d = (double)ld;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif
  return self;
}


-(BOOL)eq: (DoubleSimple *)value {
  if([value isKindOf: [DoubleLong class]]) {
    return ld == [(DoubleLong *)value lget] ? YES : NO;
  }
  else {
    BOOL result;
#ifdef INTEL
    save_pctl = fpsetprecision(FP_PC_DBL);
#endif

    result = (double)ld == [value get] ? YES : NO;

#ifdef INTEL
    fpsetprecision(save_pctl);
#endif
    return result;
  }
}

-(BOOL)ne: (DoubleSimple *)value {
  if([value isKindOf: [DoubleLong class]]) {
    return ld != [(DoubleLong *)value lget] ? YES : NO;
  }
  else {
    BOOL result;
#ifdef INTEL
    save_pctl = fpsetprecision(FP_PC_DBL);
#endif

    result = (double)ld != [value get] ? YES : NO;

#ifdef INTEL
    fpsetprecision(save_pctl);
#endif
    return result;
  }
}

-(BOOL)gt: (DoubleSimple *)value {
  if([value isKindOf: [DoubleLong class]]) {
    return ld > [(DoubleLong *)value lget] ? YES : NO;
  }
  else {
    BOOL result;
#ifdef INTEL
    save_pctl = fpsetprecision(FP_PC_DBL);
#endif

    result = (double)ld > [value get] ? YES : NO;

#ifdef INTEL
    fpsetprecision(save_pctl);
#endif
    return result;
  }
}

-(BOOL)le: (DoubleSimple *)value {
  if([value isKindOf: [DoubleLong class]]) {
    return ld <= [(DoubleLong *)value lget] ? YES : NO;
  }
  else {
    BOOL result;
#ifdef INTEL
    save_pctl = fpsetprecision(FP_PC_DBL);
#endif

    result = (double)ld <= [value get] ? YES : NO;

#ifdef INTEL
    fpsetprecision(save_pctl);
#endif
    return result;
  }
}

-(BOOL)ge: (DoubleSimple *)value {
  if([value isKindOf: [DoubleLong class]]) {
    return ld >= [(DoubleLong *)value lget] ? YES : NO;
  }
  else {
    BOOL result;
#ifdef INTEL
    save_pctl = fpsetprecision(FP_PC_DBL);
#endif

    result = (double)ld >= [value get] ? YES : NO;

#ifdef INTEL
    fpsetprecision(save_pctl);
#endif
    return result;
  }
}

-(BOOL)lt: (DoubleSimple *)value {
  if([value isKindOf: [DoubleLong class]]) {
    return ld < [(DoubleLong *)value lget] ? YES : NO;
  }
  else {
    BOOL result;
#ifdef INTEL
    save_pctl = fpsetprecision(FP_PC_DBL);
#endif

    result = (double)ld < [value get] ? YES : NO;

#ifdef INTEL
    fpsetprecision(save_pctl);
#endif
    return result;
  }
}


-(BOOL)ieq: (int)value {
  BOOL result;
#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_EXT);
#endif

  result = ld == (long double)value ? YES : NO;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif

  return result;
}

-(BOOL)ine: (int)value {
  BOOL result;
#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_EXT);
#endif

  result = ld != (long double)value ? YES : NO;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif

  return result;
}

-(BOOL)igt: (int)value {
  BOOL result;
#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_EXT);
#endif

  result = ld > (long double)value ? YES : NO;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif

  return result;
}

-(BOOL)ile: (int)value {
  BOOL result;
#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_EXT);
#endif

  result = ld <= (long double)value ? YES : NO;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif

  return result;
}

-(BOOL)ige: (int)value {
  BOOL result;
#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_EXT);
#endif

  result = ld >= (long double)value ? YES : NO;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif

  return result;
}

-(BOOL)ilt: (int)value {
  BOOL result;
#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_EXT);
#endif

  result = ld < (long double)value ? YES : NO;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif

  return result;
}


-(BOOL)ceq: (double)value {
  BOOL result;
#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_EXT);
#endif

  result = ld == (long double)value ? YES : NO;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif

  return result;
}

-(BOOL)cne: (double)value {
  BOOL result;
#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_EXT);
#endif

  result = ld != (long double)value ? YES : NO;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif

  return result;
}

-(BOOL)cgt: (double)value {
  BOOL result;
#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_EXT);
#endif

  result = ld > (long double)value ? YES : NO;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif

  return result;
}

-(BOOL)cle: (double)value {
  BOOL result;
#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_EXT);
#endif

  result = ld <= (long double)value ? YES : NO;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif

  return result;
}

-(BOOL)cge: (double)value {
  BOOL result;
#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_EXT);
#endif

  result = ld >= (long double)value ? YES : NO;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif

  return result;
}

-(BOOL)clt: (double)value {
  BOOL result;
#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_EXT);
#endif

  result = ld < (long double)value ? YES : NO;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif

  return result;
}


-(BOOL)lceq: (long double)value {
  return ld == value ? YES : NO;
}

-(BOOL)lcne: (long double)value {
  return ld != value ? YES : NO;
}

-(BOOL)lcgt: (long double)value {
  return ld > value ? YES : NO;
}

-(BOOL)lcle: (long double)value {
  return ld <= value ? YES : NO;
}

-(BOOL)lcge: (long double)value {
  return ld >= value ? YES : NO;
}

-(BOOL)lclt: (long double)value {
  return ld < value ? YES : NO;
}


-(long double)lcheckConst: (const char *)value {
  long double tmp;

#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_EXT);
#endif

  sscanf(value, "%Lf", &tmp);

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif
  return tmp;
}

@end
